-------------------------------------------------------------------------------
-- TCL2ADA                                                                   --
-- (C) Copyright 1999 ADALOG                                                 --
-- Author: J-P. Rosen                                                        --
-- Version: 1.2                                                             --
--                                                                           --
-- Rights to use, distribute or modify this procedure in any way is hereby   --
-- granted, provided this header is kept unchanged in all versions and the   --
-- associated documentation file is distributed unchanged. Additionnal       --
-- headers or documentation may be added.                                    --
-- All modifications must be properly marked as not originating from Adalog. --
-- If you make a valuable addition, please keep us informed by sending a     --
-- message to rosen.adalog@wanadoo.fr                                        --
--                                                                           --
-- ADALOG is providing training, consultancy and expertise in Ada and        --
-- related software engineering techniques. For more info about our services:--
-- ADALOG                   Tel: +33 1 41 24 31 40                           --
-- 19-21 rue du 8 mai 1945  Fax: +33 1 41 24 07 36                           --
-- 94110 ARCUEIL            E-m: rosen.adalog@wanadoo.fr                     --
-- FRANCE                   URL: http://pro.wanadoo.fr/adalog                --
--                                                                           --
-- This unit is distributed in the hope that it will be useful, but WITH-    --
-- OUT ANY WARRANTY;  without even the  implied warranty of MERCHANTABILITY  --
-- or FITNESS FOR A PARTICULAR PURPOSE.                                      --
-------------------------------------------------------------------------------

-- Usage : tcl2ada [<input-file> | - ] [<output-file> | - ]
-- A single "-" as a file name represents Standart_Input or Standard_Output.
-- if only one file name is given, it is overwritten
-- (i.e. output-file = input-file).
-- If there are no arguments (or a single "-"), arguments are defaulted to "-"
-- i.e. "tcl2ada" = "tcl2ada -" = "tcl2ada - -".
--
-- tcl2ada copies <input-file> to <output-file>, except for lines between a
-- "--tcl2ada off" comment and a "--tcl2ada on" comment.
-- if it encounters an input line of the form:
--    --tcl2ada tcl_file
-- it includes the file "tcl_file" as a properly formated Ada string.
-- Leading blanks, empty lines, and tcl comments are removed, as well as any
-- line between "#tcl2ada off" and "#tcl2ada on" comments.
-- Ascii.LF characters are inserted in place of line breaks.
-- The included file may also contain "#tcl2ada tcl-file" comments, for
-- recursively included files.

with Ada.Command_Line; use Ada.Command_Line;
with Text_IO;          use Text_IO;
with Ada.Exceptions;   use Ada.Exceptions;
with Ada.Strings.Fixed;
procedure Tcl2Ada is
   Input   : File_Type;
   Output  : File_Type;
   Line    : String (1..250);
   Length  : Natural;
   Tag     : constant String := "tcl2ada ";
   On_Tag  : constant String := "on";
   Off_Tag : constant String := "off";

   C : Character;

   -- When reading under U*x files with lines that end with CR/LF, we may get a CR
   -- as the last character. Replacing Get_Line with the following procedure gets rid
   -- of this.
   procedure Safe_Get_Line (File : in File_Type; Item : out String; Last : out Natural) is
   begin
      Get_Line (File, Item, Last);
      if Last >= Item'First and then Item (Last) = Ascii.CR then
         Last := Last - 1;
      end if;
   end Safe_Get_Line;

   procedure Safe_Get_Line (Item : out String; Last : out Natural) is
   begin
      Safe_Get_Line (Current_Input, Item, Last);
   end Safe_Get_Line;

   -- Checks that Text is a Tcl2ada command
   -- returns the argument (what follows the tag) if yes
   -- returns "" if no
   function Tcl2Ada_Arg (Text : String) return String is
      use Ada.Strings, Ada.Strings.Fixed;
   begin
      if Text'Length < Tag'Length then
         return "";
      elsif Text (Text'First..Text'First+Tag'Length-1) = Tag then
         return Trim (Text(Text'First+Tag'Length..Text'Last), Both);
      else
         return "";
      end if;
   end Tcl2Ada_Arg;

   procedure Include_File (File : String) is
      Tcl_File : File_Type;
      Tag_Check: String (1..250);
      Length   : Natural;
      Skipping : Boolean := False;
   begin
      Open (Tcl_File, In_File, File);

      --  We exit the following loop by catching End_Error rather than having
      --  a "while not End_Of_File" because if the input file is not properly formatted
      --  (i.e. the last character is not a NL), proper detection of End_Of_Line and
      --  End_Of_File may not work.
      loop
         loop
            -- Invariant: We are at the start of the line, or skipping initial spaces.
            Get (Tcl_File, C);
            if C = '#' then
               Safe_Get_Line (Tcl_File, Tag_Check, Length);
               declare
                  Arg : String := Tcl2Ada_Arg (Tag_Check(1..Length));
               begin
                  if Arg = On_Tag then
                     Skipping := False;
                  elsif Arg = Off_Tag then
                     Skipping := True;
                  elsif Arg /= "" then
                     Include_File (Arg);
                  end if;
                  if Col(Tcl_File) /= 1 then
                     Skip_Line (Tcl_File);
                  end if;
               end;
            elsif Skipping then
               Skip_Line (Tcl_File);
            elsif C /= ' ' and C /= Ascii.HT then
               exit;
            end if;
         end loop;

         Put ('"');
         loop
            case C is
               when '"' =>
                  Put ('"');
                  Put ('"');
               when Ascii.HT =>
                  Put (' ');
               when others =>
                  Put (C);
            end case;

            exit when End_Of_Line (Tcl_File);
            Get (Tcl_File, C);
         end loop;

         Put_Line  (""" & Ascii.LF &");
         Skip_Line (Tcl_File);
      end loop;

   exception
      when Name_Error =>
         Put_Line (Current_Error, "Unable to open '" & File & ''');
         Set_Exit_Status (Failure);
      when End_Error =>
         Close (Tcl_File);
      when others =>
         Put_Line (Current_Error, "Error during include");
         if Is_Open (Tcl_File) then
            Close (Tcl_File);
         end if;
         raise;
   end Include_File;

   Skipping : Boolean := False;
   procedure Copy_Line is
   -- Copy the line to output (unless we are skipping).
   -- The beginning of the line is loaded in Line.
   begin
      if Skipping then
         if Col(Current_Input) /= 1 then
            Skip_Line;
         end if;
         return;
      end if;

      -- This loop works even if the input line is longer than Line
      loop
         Put ((Line (1..Length)));
         exit when Col (Current_Input) = 1;
         Safe_Get_Line (Line, Length);
      end loop;
      New_Line;
   end Copy_Line;

   Temporary_Used : Boolean := False;
   procedure Rewrite_File is
      -- Rewrites the temporary file over the input file.
      -- Beware: here, Input is open on output, and Output on input!
   begin
      -- It is not allowed to reset a file that is the current input or
      -- the current output:
      Set_Input (Standard_Input);
      Set_Output (Standard_Output);

      Reset (Input, Out_File);
      Reset (Output, In_File);

      while not End_Of_File (Output) loop
         -- This loop works even if the input line is longer than Line
         loop
            Safe_Get_Line (Output, Line, Length);
            Put (Input, (Line (1..Length)));
            exit when Col (Output) = 1;
         end loop;
         New_Line (Input);
      end loop;
   end Rewrite_File;

   use Ada.Strings, Ada.Strings.Fixed;
begin
   if Argument_Count > 2 then
      Put_Line ("Usage : tcl2ada [<input-file> | - ] [<output-file> | - ]");
      return;
   end if;

   if Argument_Count >= 1 and then Argument(1) /= "-" then
      Open   (Input,  In_File,  Argument(1));
      Set_Input (Input);
   end if;
   if Argument_Count = 2 and then Argument(2) /= "-" then
      Create (Output, Out_File, Argument(2));
      Set_Output (Output);
   end if;
   if Argument_Count = 1 and then Argument(1) /= "-" then
      Create (Output, Out_File, "");    -- Temporary file
      Set_Output (Output);
      Temporary_Used := True;
   end if;

   while not End_Of_File (Current_Input) loop
      Safe_Get_Line (Line, Length);

      declare
         Trimmed_Line : String := Trim (Line(1..Length), Both);
      begin
         if Trimmed_Line (1..Integer'Min(2, Trimmed_Line'Last)) = "--" then
            declare
               Arg : String := Tcl2Ada_Arg (Trimmed_Line (3..Trimmed_Line'Last));
            begin
               if Arg = On_Tag then
                  if Col (Current_Input) /=1 then
                     Skip_Line;
                  end if;
                  Skipping := False;
               elsif Arg = Off_Tag then
                  if Col (Current_Input) /=1 then
                     Skip_Line;
                  end if;
                  Skipping := True;
               elsif Arg /= "" then
                  if Col (Current_Input) /=1 then
                     Skip_Line;
                  end if;
                  -- Note that the "tcl2ada off/on" and the empty string must be output here
                  -- and not from Include_file for recursive includes to work.
                  Put_Line (Line (1..Length));   -- keep the "--tcl2ada file" in the output
                  Put_Line ("--tcl2ada off");
                  Include_File (Trim (Trimmed_Line (Tag'Length+2+1..Trimmed_Line'Length), Both));
                  Put_Line ("""""");    -- This is an empty string ;-).
                  Put_Line ("--tcl2ada on");
               else
                  Copy_Line;
               end if;
            end;
         else
            Copy_Line;
         end if;
      end;
   end loop;

   if Temporary_Used then
      Rewrite_File;
   end if;

exception
   when Name_Error =>
      Put_Line (Current_Error, "Unable to find file " & Argument (1));
      Set_Exit_Status (Failure);
   when Occur : others =>
      Put (Current_Error,
         "Internal error detected, please report to rosen.adalog@wanadoo.fr");
      New_Line (Current_Error);
      Put (Current_Error, "Exception   : ");
      Put (Current_Error, Exception_Name (Occur));
      New_Line (Current_Error);
      Put (Current_Error, "Message     : ");
      Put (Current_Error, Exception_Message (Occur));
      New_Line (Current_Error);
      Set_Exit_Status (Failure);
end Tcl2Ada;
